/*------------------------------------------------------------------------------
    File        : customsp.p
    Description : Super procedure for site-specific additions
    Author(s)   : Thomas Mercer-Hursh
    Created     : Thu Jan 17 10:22:07 PST 2008
    Notes       : Use this procedure to add site-specific code.
    
NOTE: NO DEFAULT VERSION
THIS IS PROLINT VERSION, PROVIDED AS AN EXAMPLE

For simplicity, retain all procedure definitions as provided so that these
can be run by standard code at appropriate places without modification.
If the particular element does not apply, just delete the contents of the 
procedure so that it becomes a no-op, but leave the shell intact.
  ----------------------------------------------------------------------------*/
using com.cintegrity.Lib.Exception.*.
/* ***************************  Definitions  ******************************** */
                                                                   /* Classes */
define variable lobExceptionStack as class SimpleExceptionStack no-undo.
                                                               /* "Parameters */
define variable lchModuleDescriptions as character no-undo initial "ModuleDescriptions.txt".
define variable lchProgramDescriptions as character no-undo initial "ProgramDescriptions.txt".
define variable lchMenus as character no-undo initial "Menus.txt".
define variable lchMenuEntries as character no-undo initial "MenuEntries.txt".
define variable lchDynamicSets as character no-undo initial "DynamicSets.txt".
define variable lchDynamicChoices as character no-undo initial "DynamicChoices.txt".
                                                                                             /* Select one option according to the target OS */
define variable lchNewLine as character no-undo.
/*
lchNewLine = chr(10).                                             /* For Unix */
*/
lchNewLine = chr(13) + chr(10).                                /* For Windows */
                                                       
                                                               /* Package IDs */
define variable linUIModelPackageID as integer no-undo.
                                                               /* Temp-tables */
define temp-table ttPackage no-undo                   /* Package Descriptions */
    field chPackageName as character
    field chNotes as character
    index ixByName is unique primary chPackageName
    index ixByDescription chNotes
    .
define temp-table ttProgram no-undo                          /* Program Notes */
    field chProgramName as character
    field chNotes as character
    field chAlias as character
    index ixByName is unique primary chProgramName
    .
define temp-table ttMenu no-undo                                     /* Menus */
    field chMenuID as character
    field chMenuName as character
    field inMenuObjectID as integer
    index ixByID is unique primary chMenuID
    .
define temp-table ttMenuEntry no-undo                           /* Menu Items */
    field chMenuID as character
    field chOrder as character
    field chEntryText as character
    field chActionType as character
    field chActionTarget as character
    field inMenuEntryObjectID as integer
    index ixByMenuIDOrder is unique primary chMenuID chOrder
    .    
                                                              /* Dynamic Sets */
define temp-table ttDynamicSet no-undo
    field chDynamicSetID as character
    field chDynamicSetName as character
    field inDynamicSetObjectID as integer
    index ixByID is unique primary chDynamicSetID
    .
                                                           /* Dynamic Choices */
define temp-table ttDynamicChoice no-undo
    field chDynamicSetID as character
    field chOrder as character
    field chChoiceName as character
    field chActionTarget as character
    field inDynamicChoiceObjectID as integer
    index ixByDynamicIDOrder is unique primary chDynamicSetID chOrder
    .                                                               
                                                             /* Menu Packages */
define temp-table ttMenuPackage no-undo
    field chType as character
    field chPackageName as character
    field inPackageID as integer
    field inObjectID as integer
    index ixByName as primary unique chType chPackageName
    .    
/* *****************************  Functions ********************************* */
function sfCompileName returns character ( 
    input ipchProgramName as character ) 
    in super. 
/* ***************************  Main Block  ********************************* */
                    /* Do anything here to read custom data, else leave empty */
procedure spInitializeCustom:
      
  define variable iinEntry as integer no-undo.
  define variable ichName as character no-undo.
  define variable ichNotes as character  no-undo.
  define variable ichID as character no-undo.
  define variable ichOrder as character no-undo.
  define variable ichType as character no-undo.
  define variable ichTarget as character no-undo.
  define variable iinError as integer no-undo.
  define variable ichDiskFile as character no-undo.
                  /* Instantiate exception handler - leave in if used anywhere*/
  lobExceptionStack = new SimpleExceptionStack().
  lobExceptionStack:trace( "Start of Custom Preprocessing at " + string(now) ).
                                  /* Find base package for building UML menus */
  run spFindUIModelPackage (
      output linUIModelPackageID
      ).
  if lobExceptionStack:numExceptions( 1 ) gt 0
  then do:
    lobExceptionStack:pushException( 801, 1, "Fatal Error in com.cintegrity.ADBtoUML.code.codemain.p - aborting" ).
    run CleanUp.
    return.
  end. 
                                                       /* Module Descriptions */
  ichDiskFile = search( lchModuleDescriptions ).
  if ichDiskFile eq ?
  then do:
    lobExceptionStack:pushException( 802, 1, "Cannot find " + lchModuleDescriptions + " - aborting" ).
    run CleanUp.
    return.
  end.
  input from value( ichDiskFile ) no-echo.
  repeat:
    import delimiter '|' ichName ichNotes.
    create ttPackage.
    assign
        ttPackage.chPackageName = ichName
        ttPackage.chNotes = ichName + " - " + ichNotes 
        .
  end.
  input close.
                                                      /* Program Descriptions */
  ichDiskFile = search( lchProgramDescriptions ).
  if ichDiskFile eq ?
  then do:
    lobExceptionStack:pushException( 803, 1, "Cannot find " + lchProgramDescriptions + " - aborting" ).
    run CleanUp.
    return.
  end.
  input from value( ichDiskFile ) no-echo.
  repeat:
    import delimiter '|' ichName ichNotes.
    create ttProgram.
    assign
        ttProgram.chProgramName = ichName
        ttProgram.chAlias = ichName + " - " + ichNotes
        ttProgram.chNotes = ichNotes 
        .
  end.
  input close.  
                                                                     /* Menus */
  ichDiskFile = search( lchMenus ).
  if ichDiskFile eq ?
  then do:
    lobExceptionStack:pushException( 804, 1, "Cannot find " + lchMenus + " - aborting" ).
    run CleanUp.
    return.
  end.
  input from value( ichDiskFile ) no-echo.
  repeat:
    import delimiter '|' ichID ichName.
    create ttMenu.
    assign
        ttMenu.chMenuID = ichID
        ttMenu.chMenuName = ichName 
        .
  end.
  input close.                     
                                                              /* Menu Entries */
  ichDiskFile = search( lchMenuEntries ).
  if ichDiskFile eq ?
  then do:
    lobExceptionStack:pushException( 805, 1, "Cannot find " + lchMenuEntries + " - aborting" ).
    run CleanUp.
    return.
  end.
  input from value( ichDiskFile ) no-echo.
  repeat:
    import delimiter '|' ichID ichOrder ichType ichName ichTarget.
    create ttMenuEntry.
    assign
        ttMenuEntry.chMenuID = ichID
        ttMenuEntry.chOrder = ichOrder
        ttMenuEntry.chActionType = ichType
        ttMenuEntry.chEntryText = ichName
        ttMenuEntry.chActionTarget = ichTarget 
        .
  end.
  input close.
                                                              /* Dynamic Sets */
  ichDiskFile = search( lchDynamicSets ).
  if ichDiskFile eq ?
  then do:
    lobExceptionStack:pushException( 806, 1, "Cannot find " + lchDynamicSets + " - aborting" ).
    run CleanUp.
    return.
  end.
  input from value( ichDiskFile ) no-echo.
  repeat:
    import delimiter '|' ichID ichName.
    create ttDynamicSet.
    assign
        ttDynamicSet.chDynamicSetID = ichID
        ttDynamicSet.chDynamicSetName = ichName 
        .
  end.
  input close.                     
                                                           /* Dynamic Choices */
  ichDiskFile = search( lchDynamicChoices ).
  if ichDiskFile eq ?
  then do:
    lobExceptionStack:pushException( 807, 1, "Cannot find " + lchModuleDescriptions + " - aborting" ).
    run CleanUp.
    return.
  end.
  input from value( ichDiskFile ) no-echo.
  repeat:
    import delimiter '|' ichID ichOrder ichName ichTarget.
    create ttDynamicChoice.
    assign
        ttDynamicChoice.chDynamicSetID = ichID
        ttDynamicChoice.chOrder = ichOrder
        ttDynamicChoice.chChoiceName = ichName
        ttDynamicChoice.chActionTarget = ichTarget 
        .
  end.
  input close.  

end procedure.
/* *********************** Procedures *************************************** */
                                                     /* Get Package for Table */
procedure spGetTablePackage:
  define input parameter ipchTableName as character no-undo.
  define output parameter opchPackageName as character no-undo.

  if ipchTableName begins "_"
  then opchPackageName = "Metaschema".
  else if ipchTableName begins "lint_"
  then opchPackageName = "Prolint".
  else if ipchTableName begins "rtb_"
  then opchPackageName = "Roundtable".
  else opchPackageName = "Sports".
  
end procedure.
/* -------------------------------------------------------------------------- */
                                                   /* Get custom Package Note */
                                                      /* Return "" by default */                                                 
procedure spGetPackageNote:
  define input  parameter ipchPackageName as character  no-undo.
  define output parameter opchPackageNote as character  no-undo.
                                      /* Return custom note or blank */
  find ttPackage no-lock
      where ttPackage.chPackageName eq ipchPackageName
      no-error.
  if available( ttPackage )
  then opchPackageNote = ttPackage.chNotes.
  else opchPackageNote = "".
end procedure.
/* -------------------------------------------------------------------------- */
                                                   /* Get custom Program Note */
                                                      /* Return "" by default */
procedure spGetProgramNotes:
  define input  parameter ipchProgramName as character  no-undo.
  define output parameter opchProgramNote as character  no-undo.
  
  find ttProgram no-lock
      where ttProgram.chProgramName eq ipchProgramName
      no-error.
  if available( ttProgram )
  then opchProgramNote = ttProgram.chNotes.
  else do:
    opchProgramNote = "".
    lobExceptionStack:pushException( 811, 4, "No program notes found for " + ipchProgramName ).
  end.  
end procedure.
/* -------------------------------------------------------------------------- */
                                                  /* Get custom Program Alias */
                                                      /* Return "" by default */
procedure spGetProgramAlias:
  define input  parameter ipchProgramName as character  no-undo.
  define output parameter opchProgramAlias as character  no-undo.
  
  find ttProgram no-lock
      where ttProgram.chProgramName eq ipchProgramName
      no-error.
  if available( ttProgram )
  then opchProgramAlias = ttProgram.chAlias.

end procedure.
/* -------------------------------------------------------------------------- */
                                                   /* Get custom Program Path */
                                                      /* Return "" by default */                                                 
procedure spGetProgramPath:
  define input  parameter ipchProgramName as character  no-undo.
  define output parameter opchProgramPath as character  no-undo.
                                       /* Return full path name note or blank */
  opchProgramPath = search( ipchProgramName ).
  if opchProgramPath eq ?
  then opchProgramPath = "".

end procedure.
/* -------------------------------------------------------------------------- */
                               /* Additions to model after code base is built */
procedure spCustomAdditions:
  
  define variable iinMenuStructurePackageID as integer no-undo.
  define variable iinMenusPackageID as integer no-undo.
  define variable iinFunctionalUnitsPackageID as integer no-undo.
  define variable iinFunctionalUnitGroupPackageID as integer no-undo.
  define variable iinDynamicSetStructurePackageID as integer no-undo.
  define variable iinDynamicGroupPackageID as integer no-undo.
  define variable iinMenuGroupPackageID as integer no-undo.
  
  define variable iinConnectorID as integer no-undo.
  define variable iinObjectID as integer no-undo.
  define variable iinMenuGroupObjectID as integer no-undo.
  define variable iinMenuPackageID as integer no-undo.
  
  define variable ichNotes as character no-undo.
  define variable ichProgramPathName as character no-undo.
  define variable ichMenuPackage as character no-undo.
  define variable ichProgramUnitName as character no-undo.
  define variable ichCompileUnitName as character no-undo.
  define variable ichDynamicSetPackage as character no-undo.
  define variable iinDynamicSetGroupPackageID as integer no-undo.
  define variable iinDynamicSetGroupObjectID as integer no-undo.
  define variable iinDynamicSetPackageID as integer no-undo.
  
  define variable iinSlash as integer no-undo.
  define variable ichBaseName as character no-undo.

  define buffer ibfMenu for ttMenu.
  
  lobExceptionStack:trace( "Start of Custom Postprocessing at " + string(now) ).
                                        /* Create Package for Menu Structures */
  run spCreatePackage (
      input "Menu Structures",
      input linUIModelPackageID,
      output iinMenuStructurePackageID
      ).
  run spCreateObject (
      input "Package",
      input "Menu Structures",
      input iinMenuStructurePackageID,
      input 0,
      input "",
      input "",
      input "",
      output iinObjectID
      ).
/* .......................................................................... */
                                                 /* Create Packages for Menus */
  run spCreatePackage (
      input "Menus",
      input iinMenuStructurePackageID,
      output iinMenusPackageID
      ).
  run spCreateObject (
      input "Package",
      input "Menus",
      input iinMenusPackageID,
      input 0,
      input "",
      input "",
      input "",
      output iinObjectID
      ).
  if iinObjectID eq ?
  then do:
    lobExceptionStack:pushException( 821, 2, "Something really wrong in Custom Additions" ).
    return.          
  end.      
                                                                          
  for each ttMenu:
                                        /* Create/find package for menu group */
    ichMenuPackage = "Desktop".               /* Only one package for Prolint */

    run FindPackage(
        input "Menus",
        input ichMenuPackage,
        input iinMenusPackageID,
        input "",
        output iinMenuGroupPackageID,
        output iinMenuGroupObjectID
        ).
    if     iinMenuGroupObjectID eq ?
        or iinMenuGroupObjectID eq 0
    then do:
      lobExceptionStack:pushException( 822, 2, "Failure to FindPackage for " + ichMenuPackage ).
    end.
                                        /* Create package for individual menu */
    run spCreatePackage (
        input ttMenu.chMenuID + " " + ttMenu.chMenuName,
        input iinMenuGroupPackageID,
        output iinMenuPackageID
        ).
    if     iinMenuPackageID eq ?
        or iinMenuPackageID eq 0
    then do:
      lobExceptionStack:pushException( 823, 2, "Failure to Create Package for " + ttMenu.chMenuID ).
    end.
    run spCreateObject (
        input "Package",
        input ttMenu.chMenuID + " " + ttMenu.chMenuName,
        input iinMenuPackageID,
        input 0,
        input "oeMenu",
        input "",
        input ttMenu.chMenuID + " " + ttMenu.chMenuName,
        output ttMenu.inMenuObjectID
        ).  
    if     ttMenu.inMenuObjectID eq ?
        or ttMenu.inMenuObjectID eq 0
    then do:
      lobExceptionStack:pushException( 824, 2, "Failure to Create Object for " + ttMenu.chMenuID ).
    end.
                                    /* Create list of menu contents for notes */
    ichNotes = "".
    for each ttMenuEntry
        where ttMenuEntry.chMenuID = ttMenu.chMenuID
        :
      ichNotes = ichNotes + substitute( "&1. &2 ( &3=&4 )",ttMenuEntry.chOrder,ttMenuEntry.chEntryText,ttMenuEntry.chActionType,ttMenuEntry.chActionTarget) + lchNewLine.
    end.
    run spAddPackageNotes (
        input iinMenuPackageID,
        input ichNotes 
        ). 
    run spAddObjectNotes (
        input ttMenu.inMenuObjectID,
        input ichNotes 
        ). 

  end.
/* .......................................................................... */
                              /* Create Package for Dynamic Choice Structures */
  run spCreatePackage (
      input "Dynamic Choice Structures",
      input iinMenuStructurePackageID,
      output iinDynamicSetStructurePackageID
      ).
  run spCreateObject (
      input "Package",
      input "Dynamic Choice Structures",
      input iinDynamicSetStructurePackageID,
      input 0,
      input "",
      input "",
      input "",
      output iinObjectID
      ).
  if iinObjectID eq ?
  then do:
    lobExceptionStack:pushException( 831, 2, "Something really wrong in Custom Additions" ).
    return.          
  end.      
                                                                          
  for each ttDynamicSet:
                               /* Create/find package for DynamicChoice group */
    ichDynamicSetPackage = "Desktop".                 /* Only one for Prolint */

    run FindPackage(
        input "DynamicChoiceSet",
        input ichDynamicSetPackage,
        input iinDynamicSetStructurePackageID,
        input "",
        output iinDynamicSetGroupPackageID,
        output iinDynamicSetGroupObjectID
        ).
    if     iinDynamicSetGroupObjectID eq ?
        or iinDynamicSetGroupObjectID eq 0
    then do:
      lobExceptionStack:pushException( 832, 2, "Failure to FindPackage for " + ichDynamicSetPackage ).
    end.
                          /* Create package for individual Dynamic Choice Set */
    run spCreatePackage (
        input ttDynamicSet.chDynamicSetID + " " + ttDynamicSet.chDynamicSetName,
        input iinDynamicSetGroupPackageID,
        output iinDynamicSetPackageID
        ).
    if     iinDynamicSetPackageID eq ?
        or iinDynamicSetPackageID eq 0
    then do:
      lobExceptionStack:pushException( 833, 2, "Failure to Create Package for " + ttDynamicSet.chDynamicSetName ).
    end.
    run spCreateObject (
        input "Package",
        input ttDynamicSet.chDynamicSetID + " " + ttDynamicSet.chDynamicSetName,
        input iinDynamicSetPackageID,
        input 0,
        input "oeDynamicSet",
        input "",
        input ttDynamicSet.chDynamicSetID + " " + ttDynamicSet.chDynamicSetName,
        output ttDynamicSet.inDynamicSetObjectID
        ).  
    if     ttDynamicSet.inDynamicSetObjectID eq ?
        or ttDynamicSet.inDynamicSetObjectID eq 0
    then do:
      lobExceptionStack:pushException( 834, 2, "Failure to Create Object for " + ttDynamicSet.chDynamicSetName ).
    end.
                      /* Create list of Dynamic Choice Set contents for notes */
    ichNotes = "".
    for each ttDynamicChoice
        where ttDynamicChoice.chDynamicSetID eq ttDynamicChoice.chDynamicSetID
        :
      ichNotes = ichNotes + substitute( "&1. &2 ( &3 )",ttDynamicChoice.chOrder,ttDynamicChoice.chChoiceName,ttDynamicChoice.chActionTarget) + lchNewLine.
    end.
    run spAddPackageNotes (
        input iinDynamicSetPackageID,
        input ichNotes 
        ). 
    run spAddObjectNotes (
        input ttDynamicSet.inDynamicSetObjectID,
        input ichNotes 
        ). 

  end.
/* .......................................................................... */
                           /* Create Packages for Menu Entry Functional Units */
  run spCreatePackage (
      input "Functional Units",
      input iinMenuStructurePackageID, 
      output iinFunctionalUnitsPackageID
      ).
  run spCreateObject (
      input "Package",
      input "Functional Units",
      input iinFunctionalUnitsPackageID,
      input 0,
      input "",
      input "",
      input "",
      output iinObjectID
      ).
   if     iinObjectID eq ?
       or iinObjectID eq 0
   then do:
     lobExceptionStack:pushException( 841, 1, "Failure to Create Object for FunctionalUnits" ).
     return.
   end.
                       /* Create functional units for each Program menu entry */ 
   for each ttMenuEntry
       where ttMenuEntry.chActionType eq "Program"
       :
     if ttMenuEntry.chActionTarget eq "" then next. 
     iinSlash = r-index( ttMenuEntry.chActionTarget, "/" ).
     if iinSlash eq 0
     then do:
       lobExceptionStack:pushException( 842, 1, "No slash in action type for menu " + ttMenuEntry.chEntryText ).
       next.
     end.  
     ichProgramPathName = substring( ttMenuEntry.chActionTarget, 1, iinSlash - 1 ).
     
     run FindPackage(
         input "Functional Unit",
         input ichProgramPathName,
         input iinFunctionalUnitsPackageID,
         input "",
         output iinFunctionalUnitGroupPackageID,
         output iinObjectID
         ).
     if     iinObjectID eq ?
         or iinObjectID eq 0
     then do:
       lobExceptionStack:pushException( 843, 1, "Failure to FindPackage for FU package + " + ichProgramPathName ).
     end.
     else do:
       run BaseName (
          input ttMenuEntry.chActionTarget,
          output ichBaseName
          ).
       run spCreateObject (
          input "Component",
          input ttMenuEntry.chActionTarget,
          input iinFunctionalUnitGroupPackageID,
          input 0,
          input "oeFunctionalUnit",
          input ttMenuEntry.chEntryText,   /* should contain list of contents */
          input ichBaseName + " - " + ttMenuEntry.chEntryText,
          output ttMenuEntry.inMenuEntryObjectID
          ).  
       if     ttMenuEntry.inMenuEntryObjectID eq ?
           or ttMenuEntry.inMenuEntryObjectID eq 0
       then do:
         lobExceptionStack:pushException( 844, 2, "Failure to Create Object for FU + " + ichProgramPathName + "/" + ttMenuEntry.chActionTarget ).
         next.
       end.
                                                      /* Link to compile unit */
       ichCompileUnitName = sfCompileName( ttMenuEntry.chActionTarget ). 
       run spFindObject (
           input "Package",
           input "oeCompileUnit",
           input ichCompileUnitName,
           output iinObjectID
           ).
       if     iinObjectID eq ?
           or iinObjectID eq 0
       then do:
         lobExceptionStack:pushException( 845, 2, "Failure to Find Object for FU + " + ichCompileUnitName ).
         next.
       end.
       run spCreateConnector (
           input "Dependency",
           input "",
           input "",
           input "",
           input "",
           input ttMenuEntry.inMenuEntryObjectID,
           input iinObjectID,
           input "oeFunctionalLink",
           output iinConnectorID
           ).
       if     iinConnectorID eq ?
           or iinConnectorID eq 0
       then do:
         lobExceptionStack:pushException( 846, 2, "Failure to Create Connector for FU + " + ichProgramPathName + "/" + ttMenuEntry.chActionTarget + ".r" ).
         next.
       end.
     end.        
   end.
/* .......................................................................... */
                          /* Create Functional Units for each Dynnamic Choice */ 
   for each ttDynamicChoice:
     if ttDynamicChoice.chActionTarget eq "" then next. 
     iinSlash = r-index( ttDynamicChoice.chActionTarget, "/" ).
     if iinSlash eq 0
     then do:
       lobExceptionStack:pushException( 851, 1, "No slash in action type for dynamic choice " + ttDynamicChoice.chChoiceName ).
       next.
     end.  
     ichProgramPathName = substring( ttDynamicChoice.chActionTarget, 1, iinSlash - 1 ).
     
     run FindPackage(
         input "Functional Unit",
         input ichProgramPathName,
         input iinFunctionalUnitsPackageID,
         input "",
         output iinFunctionalUnitGroupPackageID,
         output iinObjectID
         ).
     if     iinObjectID eq ?
         or iinObjectID eq 0
     then do:
       lobExceptionStack:pushException( 852, 1, "Failure to FindPackage for FU package + " + ichProgramPathName ).
     end.
     else do:
       run BaseName (
          input ttDynamicChoice.chActionTarget,
          output ichBaseName
          ).
       run spCreateObject (
          input "Component",
          input ttDynamicChoice.chActionTarget,
          input iinFunctionalUnitGroupPackageID,
          input 0,
          input "oeFunctionalUnit",
          input ttDynamicChoice.chChoiceName,   
          input ichBaseName + " - " + ttDynamicChoice.chChoiceName,
          output ttDynamicChoice.inDynamicChoiceObjectID
          ).  
       if     ttDynamicChoice.inDynamicChoiceObjectID eq ?
           or ttDynamicChoice.inDynamicChoiceObjectID eq 0
       then do:
         lobExceptionStack:pushException( 853, 2, "Failure to Create Object for FU + " + ichProgramPathName + "/" + ttDynamicChoice.chActionTarget ).
         next.
       end.
                                                      /* Link to compile unit */
       ichCompileUnitName = sfCompileName( ttDynamicChoice.chActionTarget ). 
       run spFindObject (
           input "Package",
           input "oeCompileUnit",
           input ichCompileUnitName,
           output iinObjectID
           ).
       if     iinObjectID eq ?
           or iinObjectID eq 0
       then do:
         lobExceptionStack:pushException( 854, 2, "Failure to Find Object for FU + " + ichCompileUnitName ).
         next.
       end.
       run spCreateConnector (
           input "Dependency",
           input "",
           input "",
           input "",
           input "",
           input ttDynamicChoice.inDynamicChoiceObjectID,
           input iinObjectID,
           input "oeFunctionalLink",
           output iinConnectorID
           ).
       if     iinConnectorID eq ?
           or iinConnectorID eq 0
       then do:
         lobExceptionStack:pushException( 855, 2, "Failure to Create Connector for FU + " + ichProgramPathName + "/" + ttDynamicChoice.chActionTarget + ".r" ).
         next.
       end.
    end.        
  end.
/* .......................................................................... */  
                              /* Link menus and submenus and functional units */
  for each ttMenu:
    for each ttMenuEntry
        where ttMenuEntry.chMenuID eq ttMenu.chMenuID
        :
      case ttMenuEntry.chActionType:
        when "Menu"
          then do:
            find ibfMenu
                where ibfMenu.chMenuID eq ttMenuEntry.chActionTarget
                no-error.
            if available ( ibfMenu )
            then do:    
              run spCreateConnector (
                  input "ControlFlow",
                  input "",
                  input "",
                  input "",
                  input "",
                  input ttMenu.inMenuObjectID,
                  input ibfMenu.inMenuObjectID,
                  input "oeMenuLink",
                  output iinConnectorID
                  ).
             if     iinConnectorID eq ?
                 or iinConnectorID eq 0
             then do:
               lobExceptionStack:pushException( 856, 2, "Failure to Create Connector from Menu " + ttMenu.chMenuID + " to Submenu + " + ibfMenu.chMenuID ).
               next.
             end.
            end.
            else do:
              lobExceptionStack:pushException( 857, 2, "Submenu target = " + ttMenuEntry.chActionTarget + " with no corresponding menu for " + ttMenuEntry.chMenuID ).
            end.
          end.
        when "Program"
          then do:
            if      ttMenuEntry.inMenuEntryObjectID ne ?
                and ttMenuEntry.inMenuEntryObjectID ne 0
            then do:
              run spCreateConnector (
                  input "ControlFlow",
                  input "",
                  input "",
                  input "",
                  input "",
                  input ttMenu.inMenuObjectID,
                  input ttMenuEntry.inMenuEntryObjectID,
                  input "oeMenuLink",
                  output iinConnectorID
                  ).
               if     iinConnectorID eq ?
                   or iinConnectorID eq 0
               then do:
                 lobExceptionStack:pushException( 858, 2, "Failure to Create Connector from Menu " + ttMenu.chMenuID + " to Program " + ttMenuEntry.chActionTarget ).
                 next.
               end.
             end.
          end.
        when "Dynamic"
          then do:
            find ttDynamicSet
                where ttDynamicSet.chDynamicSetID eq ttMenuEntry.chActionTarget
                no-error.
            if available ( ttDynamicSet )
            then do:    
              run spCreateConnector (
                  input "ControlFlow",
                  input "",
                  input "",
                  input "",
                  input "",
                  input ttMenu.inMenuObjectID,
                  input ttDynamicSet.inDynamicSetObjectID,
                  input "oeDynamicSetLink",
                  output iinConnectorID
                  ).
             if     iinConnectorID eq ?
                 or iinConnectorID eq 0
             then do:
               lobExceptionStack:pushException( 859, 2, "Failure to Create Connector from Menu " + ttMenu.chMenuID + " to Dynamic Set + " + ttMenuEntry.chActionTarget ).
               next.
             end.
            end.
            else do:
              lobExceptionStack:pushException( 860, 2, "Dynamic Set target = " + ttMenuEntry.chActionTarget + " with no corresponding menu for " + ttMenuEntry.chMenuID ).
            end.
          end.
        otherwise
          do:
            lobExceptionStack:pushException( 861, 2, "Unexpected action type = " + ttMenuEntry.chActionType + " for " + ttMenuEntry.chMenuID ).
          end.
      end case.
    end.
  end.
/* .......................................................................... */
                                    /* Link Dynamic Sets and Functional Units */
  for each ttDynamicSet:
    for each ttDynamicChoice
        where ttDynamicChoice.chDynamicSetID eq ttDynamicSet.chDynamicSetID
        :
      if      ttDynamicChoice.inDynamicChoiceObjectID ne ?
          and ttDynamicChoice.inDynamicChoiceObjectID ne 0
      then do:
        run spCreateConnector (
            input "ControlFlow",
            input "",
            input "",
            input "",
            input "",
            input ttDynamicSet.inDynamicSetObjectID,
            input ttDynamicChoice.inDynamicChoiceObjectID,
            input "oeDynamicSetLink",
            output iinConnectorID
            ).
         if     iinConnectorID eq ?
             or iinConnectorID eq 0
         then do:
           lobExceptionStack:pushException( 865, 2, "Failure to Create Connector from Dynamic Set " + ttDynamicSet.chDynamicSetID + " to Program " + ttDynamicChoice.chActionTarget ).
           next.
         end.
       end.
    end.
  end.
/* .......................................................................... */
                                                    /* Special Postprocessing */
  /*
  run com/cintegrity/ABLtoUML/site/postprocess.p.
  */    
end procedure.
/* ************************************************************************** */
                                                                 /* Clean up */
procedure spCustomDestruct:
  delete object lobExceptionStack.
  lobExceptionStack = ?.
  empty temp-table ttProgram.
  empty temp-table ttPackage.
  empty temp-table ttMenu.
  empty temp-table ttMenuEntry.
end procedure.
/* ************************  Helper Procedures ****************************** */
                                                /* Find/Create TT for Package */
procedure FindPackage:
    define input parameter ipchType as character  no-undo.
    define input parameter ipchPackageName as character no-undo.
    define input parameter ipinParentPackageID as integer no-undo.
    define input parameter ipchStereotype as character no-undo.
    define output parameter opinPackageID as integer no-undo.
    define output parameter opinObjectID as integer no-undo.
  
  define variable ichPackageParentName as character  no-undo.
  define variable ichPackageBaseName as character no-undo.
  define variable iinLocalParentPackageID as integer no-undo.
  define variable iinLocalParentObjectID as integer no-undo.
  define variable iinLocalPackageID as integer no-undo.
  define variable iinLocalObjectID as integer no-undo.
  define variable ichNotes as character  no-undo.
  
  find ttMenuPackage
      where ttMenuPackage.chType eq ipchType 
      and   ttMenuPackage.chPackage eq ipchPackageName 
      no-error.

  if not available( ttMenuPackage )
  then do:
    run ParsePackage (
        input ipchPackageName,
        output ichPackageParentName,
        output ichPackageBaseName
        ).
    if ichPackageParentName ne ""
    then do:
      run FindPackage (
          input ipchType,
          input ichPackageParentName,
          input ipinParentPackageID,
          input "",
          output iinLocalParentPackageID,
          output iinLocalParentObjectID
          ).
    end.
    else do:
      iinLocalParentPackageID = ipinParentPackageID.
    end.
                                                            /* Create package */
    run spCreatePackage (
        input ipchPackageName,
        input iinLocalParentPackageID,
        output iinLocalPackageID
        ).
     if iinLocalPackageID eq ?
     then do:
       lobExceptionStack:pushException( 871, 2, "Failure to CreatePackage for " + ipchPackageName ).
       next.
     end.
                                            /* Check for custom package Notes */
    run spGetPackageNote (
        input ipchPackageName,
        output ichNotes
        ).
    if ichNotes ne ""
    then do:
      run spAddPackageNotes (
          input iinLocalPackageID,
          input ichNotes
          ).
    end.
                                               /* Create corresponding object */
    run spCreateObject (
        input "Package",
        input ipchPackageName,
        input iinLocalPackageID,
        input 0,
        input ipchStereotype,
        input "",
        input "",
        output iinLocalObjectID
        ).    
     if iinLocalObjectID eq ?
     then do:
       lobExceptionStack:pushException( 872, 2, "Failure to CreateObject for " + ipchPackageName ).
       next.
     end.
                                                   /* Create Temp Table entry */
    create ttMenuPackage.
    assign
        ttMenuPackage.chType = ipchType
        ttMenuPackage.chPackageName = ipchPackageName
        ttMenuPackage.inPackageID = iinLocalPackageID
        ttMenuPackage.inObjectID = iinLocalObjectID
        .
  end.  
  assign
      opinPackageID = ttMenuPackage.inPackageID
      opinObjectID = ttMenuPackage.inObjectID
      .
end procedure.
/* -------------------------------------------------------------------------- */
procedure ParsePackage:
  define input  parameter ipchPackageFullName as character  no-undo.
  define output parameter opchPackageParent as character  no-undo.
  define output parameter opchPackageBaseName as character  no-undo.
  
  define variable iinSlash as integer    no-undo.
  
  iinSlash = r-index(ipchPackageFullName,"/").
  if iinSlash eq 0
  then iinSlash = r-index(ipchPackageFullName,"~\").
  if iinSlash eq 0
  then
    assign
        opchPackageParent = ""
        opchPackageBaseName = ipchPackageFullName
        .
  else
    assign
        opchPackageParent = substring(ipchPackageFullName, 1, iinSlash - 1 )
        opchPackageBaseName = substring(ipchPackageFullName, iinSlash + 1 )
        .
end procedure.    
/* -------------------------------------------------------------------------- */
procedure BaseName:
  define input parameter ipchFullName as character no-undo.
  define output parameter opchBaseName as character no-undo.
  
  define variable iinSlash as integer no-undo.
  
  iinSlash = r-index( ipchFullName, "/").
  if iinSlash gt 0
  then opchBaseName = substring( ipchFullName, iinSlash + 1 ).
  else opchBaseName = ipchFullName.
  if num-entries( opchBaseName, "." ) gt 0
  then opchBaseName = entry(1, opchBaseName, "." ).
end procedure.     
/************************** End of Program ************************************/